import { world, system, Player } from "@minecraft/server";
import * as vehicleArrays from "../arrays/vehicle_arrays.js"

export function vehiclesMainComponents(vehicle) {
	const vehicleOnList = vehicleArrays.vehicleTypesList.includes(vehicle.typeId);
	if (!vehicleOnList) return;
	const rule = vehicleArrays.vehicleTypes[vehicle.typeId];
	if (!rule) return;
	const engine = vehicle.getProperty(rule.properties.engine);
	const hasRider = vehicle.getProperty(rule.properties.hasRider);
	const fuel = vehicle.getProperty(rule.properties.fuelAmount);
	updateFuelScoreboard(vehicle,rule.properties.fuelAmount, fuel);
	const velocity = vehicle.getProperty(rule.properties.movementVelocity);
	updateVelocityScoreboard(vehicle,rule.properties.movementVelocity, velocity);
	if (engine=="unpowered"&&velocity!=0.00) {
		applyFriction(rule,vehicle,velocity);
	}
	if (engine=="powered") {
		handleMotorbikeMovement(rule,vehicle,hasRider,velocity);
	}
};
function updateFuelScoreboard(entity,prop,value) {
	if (value==null) return;
	const indicator = Math.floor(value*100);
	entity.runCommand(`scoreboard objectives add ${prop} dummy ${prop}`);
	entity.runCommand(`scoreboard players set @s[has_property={${prop}=${value}}] ${prop} ${indicator}`);
};
function updateVelocityScoreboard(entity,prop,value) {
	if (value==null) return;
	const indicator = Math.round(Math.abs(value)*100);
	entity.runCommand(`scoreboard objectives add ${prop} dummy ${prop}`);
	entity.runCommand(`scoreboard players set @s[has_property={${prop}=${value}}] ${prop} ${indicator}`);
};
function handleMotorbikeMovement(rule,vehicle,hasRider,velocity) {
	const riders = vehicle.getComponent("rideable").getRiders();
	const player = riders.find(r=>r instanceof Player);
	vehicle.runCommand(`titleraw @a[r=1.5] actionbar {"rawtext":[{"text":"§e"},{"translate":"action.interact.vehicle.fuel_amount"},{"text":": "},{"score":{"name":"@s","objective":"${rule.properties.fuelAmount}"}},{"text":" "},{"translate":"action.interact.vehicle.movement_velocity"},{"text":": "},{"score":{"name":"@s","objective":"${rule.properties.movementVelocity}"}}]}`);
	if (hasRider&&player) {
		player.playAnimation("animation.motorbike.player",{stopExpression: "!q.is_riding||v.is_first_person"});
		const movement = player.inputInfo?.getMovementVector?.();
		if (!movement) return;
		const normalizedX = Math.abs(movement.x)>=rule.vehicleInputs.threshold?(movement.x>0?1:-1):0;
		const normalizedY = Math.abs(movement.y)>=rule.vehicleInputs.threshold?(movement.y>0?1:-1):0;
		const key = `${normalizedX},${normalizedY}`;
		const inputType = rule.vehicleInputs.inputDirections[key]??"noAcceleration";
		const targetAccel = rule.vehicleInputs[inputType];
		adjustVelocityWithInput(rule,vehicle,velocity,targetAccel);
		applyImpulse(rule,vehicle,velocity,vehicle.isInWater?rule.vehicleInputs.movementOnWater:vehicle.isOnGround?rule.vehicleInputs.movementOnGround:rule.vehicleInputs.movementOnAir);
	}
	else {
		applyFriction(rule,vehicle,velocity);
	}
};
function adjustVelocityWithInput(rule,vehicle,currentVelocity,targetAccel) {
	// Direction switching: decelerate to 0 before switching
	if (targetAccel>rule.vehicleInputs.noAcceleration&&currentVelocity<0) {
		// Was moving backward, now trying to go forward
		const newVelocity = Math.min(0,currentVelocity+0.006);
		vehicle.setProperty(rule.properties.movementVelocity,newVelocity);
		return;
	}
	if (targetAccel<rule.vehicleInputs.noAcceleration&&currentVelocity>0) {
		// Was moving forward, now trying to go backward
		const newVelocity = Math.max(0,currentVelocity-0.006);
		vehicle.setProperty(rule.properties.movementVelocity,newVelocity);
		return;
	}
	if (targetAccel>rule.vehicleInputs.noAcceleration) {
		// Forward acceleration curve
		for (const [min,max,accel] of rule.velocityProperties.forwardAcceleration) {
			if (currentVelocity >= min && currentVelocity < max) {
				let next = currentVelocity+accel*targetAccel;
				if (vehicle.isInWater&&next>rule.maxVelocityInWater) {
					next = Math.max(next-0.005,rule.maxVelocityInWater);
				}
				next = Math.min(next,rule.maxDefaultVelocity);
				vehicle.setProperty(rule.properties.movementVelocity,next);
				return;
			}
		}
	}
	else if (targetAccel<rule.vehicleInputs.noAcceleration) {
		// Backward acceleration curve
		const absVelocity = Math.abs(currentVelocity);
		for (const [min,max,accel] of rule.velocityProperties.backwardAcceleration) {
			if (absVelocity >= min && absVelocity < max) {
				const next = Math.max(currentVelocity+accel*targetAccel,rule.minDefaultVelocity);
				vehicle.setProperty(rule.properties.movementVelocity,next);
				return;
			}
		}
	}
	else {
		for (const [min,max,decel] of rule.velocityProperties.deceleration) {
			if (currentVelocity>=min&&currentVelocity<max) {
				const next = Math.max(0,currentVelocity-decel);
				vehicle.setProperty(rule.properties.movementVelocity,next);
				return;
			}
		}
		// No input – decelerate naturally
		let decay = rule.velocityProperties.defaultDecay;
		if (currentVelocity>=rule.velocityProperties.driving.velocity) decay = rule.velocityProperties.driving.decay; // Slower decay when driving
		if (currentVelocity>=rule.velocityProperties.speeding.velocity) decay = rule.velocityProperties.speeding.decay; // Even slower decay when speeding
		const newVelocity = currentVelocity > 0
			? Math.max(0, currentVelocity-decay)
			: Math.min(0, currentVelocity+decay); // Slower decay for reverse
		vehicle.setProperty(rule.properties.movementVelocity,newVelocity);
	}
};
function applyFriction(rule,vehicle,velocity) {
	if (Math.abs(velocity)<0.02) return vehicle.setProperty(rule.properties.movementVelocity,0);
	const decay = velocity>0?0.002:0.001;
	const newVelocity = velocity>0
		? Math.max(0,velocity-decay)
		: Math.min(0,velocity+decay);
	vehicle.setProperty(rule.properties.movementVelocity,newVelocity);
	applyImpulse(rule,vehicle,velocity,vehicle.isInWater?rule.vehicleInputs.movementOnWater:vehicle.isOnGround?rule.vehicleInputs.movementOnGround:rule.vehicleInputs.movementOnAir);
};
function applyImpulse(rule,vehicle,velocity,scale=rule.vehicleInputs.movementOnDefault) {
	const dir = vehicle.getViewDirection();
	vehicle.applyImpulse({
		x: dir.x*velocity*scale,
		y: 0,
		z: dir.z*velocity*scale
	});
};

function handleVehicleWASDInputControl(rule,dragon,player,navigationState,isOnGround) {
	const movement = player.inputInfo.getMovementVector();
	if (!movement) return;
	const normalizedX = Math.abs(movement.x) >= rule.threshold ? (movement.x > 0 ? 1 : -1) : 0;
	const normalizedY = Math.abs(movement.y) >= rule.threshold ? (movement.y > 0 ? 1 : -1) : 0;
	const inputKey = `${normalizedX},${normalizedY}`;
	const inputType = rule.inputDirections[inputKey] ?? "noAcceleration";
	const acceleration = rule[inputType]??0;
	// Get forward direction
	const forward = dragon.getViewDirection();
	// Calculate right vector (perpendicular to forward vector on XZ plane)
	const right = { x: forward.z, y: 0, z: -forward.x };
	// Calculate final direction vector from input
	let moveDir = { x: 0, y: 0, z: 0 };
	if (normalizedY !== 0) {
		moveDir.x += forward.x * normalizedY;
		moveDir.z += forward.z * normalizedY;
	}
	if (normalizedX !== 0) {
		moveDir.x += right.x * normalizedX;
		moveDir.z += right.z * normalizedX;
	}
	// Normalize moveDir to prevent faster diagonals
	const magnitude = Math.hypot(moveDir.x, moveDir.z);
	if (magnitude > 0) {
		moveDir.x /= magnitude;
		moveDir.z /= magnitude;
	}
	// Apply scale depending on ground status
	const scale = dragon.isInWater?rule.movementOnWater:isOnGround?rule.movementOnGround:rule.movementOnAir;
	// Final impulse
	const pitch = player.getRotation().x;
	let altitude = 0;
	if (pitch<-22.5) {
		const upPitch = Math.max(Math.min(pitch,-22.5),-90);
		const t = (upPitch+22.5)/(-90+22.5);
		altitude = lerp(0.015,0.05,t);
	}
	else if (pitch>35) {
		const downPitch = Math.min(Math.max(pitch,35),90);
		const t = (downPitch-22.5)/(90-35);
		altitude = lerp(-0.015,-0.03,t);
	}
	else {
		altitude = -dragon.getVelocity().y;
	}
	if (pitch<-22.5&&navigationState!="air_controlled") dragon.triggerEvent("minecraft:dragon_air_controlled");
	if (pitch>35&&isOnGround&&navigationState!="grounded") dragon.triggerEvent("minecraft:dragon_grounded");
	dragon.applyImpulse({
		x: moveDir.x*acceleration*scale,
		y: altitude,
		z: moveDir.z*acceleration*scale
	});
	//player.sendMessage(`Dragon grounded: ${isOnGround}`);
	//notice: dragon is on ground but somehow pitch won't do shit until it moves.
	/*Note to self:
		Split Functions of Pitch, Ground & Air Controls.
	*/
};

function lerp(a,b,t) {
	return a+(b-a)*t;
};